import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Random;


public class MaxMinAS {
	/** dostupne stock duljine */
	private int[] stockSizes;
	/** tražene item duljine */
	private int[] itemSizes;
	/** dostupan broj stockova za svaku duljinu, negativan ako nije ograničeno */
	private int[] stockNumbers;
	/** tražen broj itema za svaku duljinu */
	private int[] itemNumbers;
	
	/** parametri algoritma */
	private Parameters param;
	
	/** broj mrava, odnosno rješenja */
	private int numOfants;
	/** globalno najbolji mrav */
	private Solution bestAnt;
	/** tragovi među itemSizes, zasebna tablica za svaki stock size */
	private double[][][] itemTrails;
	/** suma tragova itema za pojedine duljine */
	private double[] sumItemTrails;
	/** tragovi za pojedine stockSizes */
	private double[] stockTrails;
	/** vjerojatnost odabira odreðenog stocka */
	private double[] stockProbabilities;
	/** broj trenutno nerasposloženih itema za nekog mrava */
	private int[] itemsAvailable;
	/** označava da li se item još može izrezati iz stocka */
	private boolean[] isItemUsable;
	/** označava da li se trenutni stock iskoristio za pojedini item */
	private boolean[] isItemPresent;
	/** broj trenutno dostupnih stockova za nekog mrava */
	private int[] stocksAvailable;
	/** polje koje predstavlja broj pojavljivanja određenog iteme unutar nekog stocka */
	private int[] itemNumOfOccurence;
	/** relativna iskorištenost svih stockova iste duljine unutar rješenja */
	private double[] relativeStockUsage;
	/** koliko se često pojavljuje određeni stock u rješenju */
	private int[] stockNumOfOccurence;

	/**
	 * Constructor koji gradi novi ACO objekt i inicijalizira stock duljine, 
	 * item duljine, broj stockova, broj itema, generatos slučajnih brojeva 
	 * te parametre algoritma. Stvara polja za tragove i postavlja elemente
	 * na početne vrijednosti, stvara novo polje mrava veličine broja traženih 
	 * itema, te polje trenutno dostupnih stockova i neposloženih itema.
	 * @param stockSizes dostupne stock duljine
	 * @param itemSizes tražene item duljine
	 * @param stockNumbers dostupan broj stockova za svaku duljinu, negativan
	 * 		ako nije ograničeno
	 * @param itemNumbers tražen broj itema za svaku duljinu
	 * @param param parametri algoritma
	 * @param rand generator slučajnih brojeva
	 */
	public MaxMinAS(int[] stockSizes, int[] itemSizes, int[] stockNumbers,
			int[] itemNumbers, Parameters param, Random rand) {
		this.stockSizes = stockSizes;
		this.itemSizes = itemSizes;
		this.stockNumbers = stockNumbers;
		this.itemNumbers = itemNumbers;
		this.param = param;
		
		this.stocksAvailable = new int[this.stockNumbers.length];
		this.itemsAvailable = new int[this.itemNumbers.length];
		this.isItemUsable = new boolean[this.itemNumbers.length];
		this.isItemPresent = new boolean[this.itemNumbers.length];
		
		for (int i = 1; i < this.itemNumbers.length; i++) {
			this.numOfants += this.itemNumbers[i];
			if (this.numOfants > param.maxAnts) {
				this.numOfants = param.maxAnts;
				break;
			}
		}
		
		this.stockTrails = new double[this.stockSizes.length];
		this.itemTrails = new double[this.stockSizes.length]
		                        [this.itemSizes.length][this.itemSizes.length];
		this.sumItemTrails = new double[this.itemSizes.length];
		
		this.itemNumOfOccurence = new int[this.itemSizes.length];
		this.stockNumOfOccurence = new int[this.stockSizes.length];
		this.relativeStockUsage = new double[this.stockSizes.length];
		
		this.stockProbabilities = new double[this.stockSizes.length];
	}
	
	/**
	 * Metoda koja pronalazi rješenje problema zadanog pri instanciranju objekta. 
	 * @param rand generator slučajnih brojeva
	 * @return najbolje rješenje
	 */
	public Solution findSolution(Random rand) {
		//postavljanje početnih vrijednosti tragova
		double effMinTrail = 3*this.param.minTrail;
		for (int stockNum = 0; stockNum < stockSizes.length; stockNum++) {
			this.stockTrails[stockNum] = effMinTrail;
			
			for (int rowNum = 0; rowNum < itemSizes.length; rowNum++) {
				Arrays.fill(this.itemTrails[stockNum][rowNum], 2.5*this.param.minTrail);
			}
		}
		//pokreni algoritam
		long startTime = System.currentTimeMillis();
		bestAnt = new Solution();
		
		doWalk(bestAnt, rand);
		Solution iterBestAnt;
		Solution newAnt;
		for (int iter = 0; iter < param.maxIter; iter++) {
			if (System.currentTimeMillis() - startTime > param.maxTime*1000) break;
			iterBestAnt = null;
			for (int antNum = 0; antNum < numOfants; antNum++) {
				newAnt = new Solution();
				doWalk(newAnt, rand);
				
				if (iterBestAnt == null 
						|| iterBestAnt.getFitness() < newAnt.getFitness()) {
					iterBestAnt = newAnt;
					if (bestAnt.getFitness() < newAnt.getFitness()) {
						bestAnt = newAnt;
					}
				}
			}
			
			if (iter%param.gama == 0) {
				updateTrails(bestAnt); 
			}
			else {
				updateTrails(iterBestAnt);
			}
		}
		return bestAnt;
	}
	
	/**
	 * Metoda koja pronalazi jedno rješnenje. Obavlja šetnju dobivenog mrava. 
	 * @param ant mrav za kojeg treba obaviti šetnju
	 * @param rand generator slučajnih brojeva
	 */
	public void doWalk(Solution ant, Random rand) {
		System.arraycopy(stockNumbers, 0, stocksAvailable, 0, stockNumbers.length);
		System.arraycopy(itemNumbers, 0, itemsAvailable, 0, itemNumbers.length);
		double fitness = 0;
		
		int itemsDemanded = itemsAvailable[0];
		for (int i = 1; i < itemsAvailable.length; i++) {
			itemsDemanded += itemsAvailable[i];
		}
		boolean recalcStockProb = true;
		
		int itemTypesUsable;
		while(itemsDemanded > 0) {
			if (recalcStockProb) {
				calcStockProbabilities();
				recalcStockProb = false;
			}
			
			//odaberi stock
			int chosenStockNum = chooseStock(rand);
			assert chosenStockNum >= 0 : "chosenStockNum < 0";
			UsedStock newStock = new UsedStock(stockSizes[chosenStockNum]);
			
			Arrays.fill(isItemPresent, false);
			//prvi item se bira slučajno
			itemTypesUsable = 1;
			for (int i = 0; i < itemsAvailable.length; i++) {
				if (itemsAvailable[i] >= 0 && itemSizes[i] < newStock.getRemainder()) {
					itemTypesUsable += 1;
					isItemUsable[i] = true;
				} else {
					isItemUsable[i] = false;
				}
			}
			itemTypesUsable -= 1;
			if (itemTypesUsable != 0) {
				int roll = rand.nextInt(itemsAvailable.length);
				while (!isItemUsable[roll]) {
					roll = rand.nextInt(itemsAvailable.length);
				}
				boolean uspjesno = newStock.addItem(itemSizes[roll], roll, param.sawThickness);
				assert uspjesno : "Nije uspjelo dodavanje prvog itema!";
				isItemPresent[roll] = true;
				itemsAvailable[roll] -= 1;
				itemsDemanded -= 1;
				
				if (stocksAvailable[chosenStockNum] > 0) {
					stocksAvailable[chosenStockNum] -= 1;
				}
			} else {
				stocksAvailable[chosenStockNum] = 0;
			}
			if (stocksAvailable[chosenStockNum] == 0) {
				recalcStockProb = true;
			}
			//popunjavanje stocka
			while(itemsDemanded > 0 && itemTypesUsable >= 1) {
				itemTypesUsable = 1;
				for (int i = 0; i < itemsAvailable.length; i++) {
					if (itemsAvailable[i] >= 0 && itemSizes[i] < newStock.getRemainder()) {
						itemTypesUsable += 1;
						isItemUsable[i] = true;
					} else {
						isItemUsable[i] = false;
					}
				}
				itemTypesUsable -= 1;
				if (itemTypesUsable == 0) break;
				
				//izračunaj sumu tragova i odaberi item
				int chosenItemNum = chooseItem(chosenStockNum, rand);
				
				assert chosenItemNum >= 0 : "chosenItemNum < 0";
				
				boolean uspjelo = newStock.addItem(itemSizes[chosenItemNum], chosenItemNum, param.sawThickness);
				assert uspjelo : "Nije uspjelo dodavanje itema!";
				isItemPresent[chosenItemNum] = true;
				itemsAvailable[chosenItemNum] -= 1;
				itemsDemanded -= 1;
			}
			if (!newStock.isEmpty()) {
				if (newStock.getRemainder() > param.acceptableRemainder) {
					fitness += Math.pow(1 - 
						(double)newStock.getWaste()/newStock.getStockSize(),
						param.fitnessExponentK);
					
					
				} else {
					fitness += 1;
				}
				ant.add(newStock);
			}
		}
		fitness /= ant.size();
		ant.setFitness(fitness);
	}
	
	/**
	 * Metoda koja služi za odabir stocka. Biranje se vrši generiranjem slučajnog
	 * broja te odabirom uzimajući u obzir definirane vjerojatnosti pojedinih
	 * stockova. 
	 * @return broj odabranog stocka
	 */
	private int chooseStock(Random rand) {
		double roll = rand.nextDouble();
		double cumulativeProb = 0;
		for (int stockNum = 0; stockNum < stockProbabilities.length; stockNum++) {
			cumulativeProb += stockProbabilities[stockNum];
			if (roll < cumulativeProb) return stockNum;
		}
		
		return stockProbabilities.length - 1;
	}
	
	/**
	 * Metoda koja služi za odabir itema. Biranje se vrši temeljem izračunatih
	 * vjerojatnosti odabira pojedinih itema s obzirom na trenutne iteme u
	 * planu rezanja stocka i preostali kapacitet. 
	 * @param chosenStockNum broj stocka za kojeg se bira item
	 * @param rand generator slučajnih brojeva
	 * @return broj odabranog itema
	 */
	private int chooseItem(int chosenStockNum, Random rand) {
		double totalSumItemTrails = 0;
		Arrays.fill(sumItemTrails, 0);
		for (int i = 0; i < itemsAvailable.length; i++) {
			if (isItemUsable[i]) {
				for (int j = 0; j < i; j++) {
					if (isItemPresent[j]) {
						totalSumItemTrails += itemTrails[chosenStockNum][i][j];
						sumItemTrails[i] += itemTrails[chosenStockNum][i][j];
					}
				}
				for (int j = i; j < itemsAvailable.length; j++) {
					if (isItemPresent[j]) {
						totalSumItemTrails += itemTrails[chosenStockNum][i][j];
						sumItemTrails[i] += itemTrails[chosenStockNum][i][j];
					}
				}
			}
		}
		
		double roll = rand.nextDouble();
		double cumulativeProb = 0;
		for (int itemNum = 0; itemNum < sumItemTrails.length; itemNum++) {
			cumulativeProb += sumItemTrails[itemNum] / totalSumItemTrails;
			if (roll < cumulativeProb) return itemNum;
		}
		
		return -1;
	}
	
	/**
	 * Metoda koja osvježava tragove prema dobivenom mravu. 
	 * @param ant mrav koji ostavlja tragove
	 */
	private void updateTrails(Solution ant) {
		//ažuriraj tragove prema najboljem rješenju
		Arrays.fill(relativeStockUsage, 0);
		Arrays.fill(stockNumOfOccurence, 0);
		for(UsedStock stock : ant) {
			int stockNumber;
			boolean found = false;
			for (stockNumber = 0; stockNumber < stockSizes.length; stockNumber++) {
				if (stock.getStockSize() == stockSizes[stockNumber]) {
					found = true;
					break;
				}
			}
			assert found : "Nije nađen stock veličine " + stock.getStockSize();
			
			if (stock.getRemainder() > param.acceptableRemainder) {
				//izračunaj za stock
				relativeStockUsage[stockNumber] += Math.pow(1 - 
						(double) stock.getWaste() / stock.getStockSize(),
						this.param.fitnessExponentK);
			} else {
				relativeStockUsage[stockNumber] += 1;
			}
			stockNumOfOccurence[stockNumber] += 1;
			
			//izračunaj za iteme unutar stocka
			Arrays.fill(itemNumOfOccurence, 0);
			for (int i = 0; i < stock.numOfItems(); i++) {
				itemNumOfOccurence[stock.getItemNumber(i)] += 1;
			}
			for (int i = 0; i < itemNumOfOccurence.length; i++) {
				if (itemNumOfOccurence[i] >= 1) {
					if (itemNumOfOccurence[i] >= 2) {
						itemTrails[stockNumber][i][i] += ant.getFitness();
					}
					for (int j = i + 1; j < itemNumOfOccurence.length; j++) {
						if (itemNumOfOccurence[j] >= 1) {
							itemTrails[stockNumber][i][j] += ant.getFitness();
						}
					}
				}
			}
		}
		
		//ispari
		for (int stockNum = 0; stockNum < stockSizes.length; stockNum++) {
			if (stockNumOfOccurence[stockNum] != 0) { 
				stockTrails[stockNum] = (stockTrails[stockNum] + 
						relativeStockUsage[stockNum] / stockNumOfOccurence[stockNum]) 
						* param.ro;
			} else {
				stockTrails[stockNum] = stockTrails[stockNum] * param.ro;
			}
			if (stockTrails[stockNum] < param.minTrail) {
				stockTrails[stockNum] = param.minTrail;
			}
			for (int i = 0; i < itemSizes.length; i++) {
				for (int j = i; j < itemSizes.length; j++) {
					itemTrails[stockNum][i][j] *= param.ro;
					if (itemTrails[stockNum][i][j] < param.minTrail) {
						itemTrails[stockNum][i][j] = param.minTrail;
					}
				}
			}
		}
	}
	
	/**
	 * Metoda koja računa vjerojatnosti odabira određenog stocka na temelju 
	 * tragova i heurističke informacije za raspoložive stockove (one kojih još
	 * ima i u koji stane barem jedan item). 
	 */
	private void calcStockProbabilities() {
		double sumTrailHeuristic = 0;
		for (int stockNum = 0; stockNum < stocksAvailable.length; stockNum++) {
			if (this.stocksAvailable[stockNum] > 0) {
				sumTrailHeuristic += this.stockTrails[stockNum]
						* this.param.stockHeuristicFavorability;
			} else if (this.stocksAvailable[stockNum] < 0) {
				sumTrailHeuristic += this.stockTrails[stockNum];
			}
		}
		assert sumTrailHeuristic != 0 : "Division by 0";
		
		for (int stockNum = 0; stockNum < stocksAvailable.length; stockNum++) {
			double probability;
			if (stocksAvailable[stockNum] > 0) {
				probability = stockTrails[stockNum] * param.stockHeuristicFavorability 
								/ sumTrailHeuristic;
			} else if (stocksAvailable[stockNum] < 0){
				probability = stockTrails[stockNum] / sumTrailHeuristic;
			} else {
				probability = 0;
			}
			stockProbabilities[stockNum] = probability;
		}
	}
	
	/**
	 * Metoda koja se poziva pokretanjem programa. 
	 * @param args argumenti komandne linije, očekuje se ime zadatka
	 */
	public static void main(String[] args) {
		List<Integer> stockSizesList = new ArrayList<Integer>();
		List<Integer> stockNumbersList = new ArrayList<Integer>();
		List<Integer> itemSizesList = new ArrayList<Integer>();
		List<Integer> itemNumbersList = new ArrayList<Integer>();
		
		BufferedReader reader = null;
		
		try {
			reader = new BufferedReader(new FileReader("algoritmi\\" + args[0] + ".in"));
		} catch (FileNotFoundException e) {
			System.err.println("Nije nađena ulazna datoteka!");
			e.printStackTrace();
			System.exit(1);
		}
		
		double maxTime = 0;
		int sawThickness = 0;
		int acceptableRemainder = 0;
		
		String redak;
		try {
			String[] razdvojenRedak;
			
			razdvojenRedak = reader.readLine().trim().split("\\s");
			maxTime = Double.parseDouble(razdvojenRedak[0]);
			sawThickness = Integer.parseInt(razdvojenRedak[1]);
			acceptableRemainder = Integer.parseInt(razdvojenRedak[2]);
			
			while ((redak = reader.readLine()) != null && !redak.isEmpty() && !redak.startsWith("--")) {
				razdvojenRedak = redak.split("\\s");
				stockSizesList.add(Integer.parseInt(razdvojenRedak[0]));
				stockNumbersList.add(Integer.parseInt(razdvojenRedak[1]));
			}
			
			while ((redak = reader.readLine()) != null && !redak.isEmpty()) {
				razdvojenRedak = redak.split("\\s");
				itemSizesList.add(Integer.parseInt(razdvojenRedak[0]));
				itemNumbersList.add(Integer.parseInt(razdvojenRedak[1]));
			}
		} catch (IOException e) {
			System.err.println("Greska pri citanju ulazne datoteke");
			e.printStackTrace();
			System.exit(1);
		}
		try {
			reader.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		int[] stockSizes = new int[stockSizesList.size()];
		int[] stockNumbers = new int[stockSizesList.size()];
		int[] itemSizes = new int[itemSizesList.size()];
		int[] itemNumbers = new int[itemSizesList.size()];
		
		for (int i = 0; i < stockSizesList.size(); i++) {
			stockSizes[i] = stockSizesList.get(i);
			stockNumbers[i] = stockNumbersList.get(i);
		}
		for (int i = 0; i < itemSizesList.size(); i++) {
			itemSizes[i] = itemSizesList.get(i);
			itemNumbers[i] = itemNumbersList.get(i);
		}
		
		double ro = 0.95;
		int gama = 5;
		double minTrail = 0.5;
		double stockHeuristicFavorability = 1.5;
		int maxIter = 100;
		double k = 2;
		int maxAnts = 50;
		
		try {
			reader = new BufferedReader(new FileReader("algoritmi\\MaxMinAS.config"));
		} catch (FileNotFoundException e) {
			System.err.println("Nije nađena konfiguracijska datoteka!");
			e.printStackTrace();
			System.exit(1);
		}
		try {
			String[] razdvojenRedak;
			while ((redak = reader.readLine()) != null && !redak.isEmpty()) {
				razdvojenRedak = redak.split("\\s*=\\s*");
				if (razdvojenRedak[0].equalsIgnoreCase("evaporation constant")) {
					ro = Double.parseDouble(razdvojenRedak[1]);
				} else if (razdvojenRedak[0].equalsIgnoreCase("gama")) {
					gama = Integer.parseInt(razdvojenRedak[1]);
				} else if (razdvojenRedak[0].equalsIgnoreCase("minimum trail")) {
					minTrail = Double.parseDouble(razdvojenRedak[1]);
				} else if (razdvojenRedak[0].equalsIgnoreCase("max iterations")) {
					maxIter = Integer.parseInt(razdvojenRedak[1]);
				} else if (razdvojenRedak[0].equalsIgnoreCase("k")) {
					k = Double.parseDouble(razdvojenRedak[1]);
				} else if (razdvojenRedak[0].equalsIgnoreCase("stock heuristic favorability")) {
					stockHeuristicFavorability = Double.parseDouble(razdvojenRedak[1]);
				} else if (razdvojenRedak[0].equalsIgnoreCase("max ants")) {
					maxAnts = Integer.parseInt(razdvojenRedak[1]);
				}
			}
		} catch (IOException e) {
			System.err.println("Greska pri citanju konfiguracijske datoteke");
			e.printStackTrace();
			System.exit(1);
		}
		try {
			reader.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		Parameters param = new Parameters(sawThickness, ro, gama, minTrail, 
				stockHeuristicFavorability, maxIter, k, acceptableRemainder, maxTime, maxAnts);
		Random rand = new Random();
		MaxMinAS algoritam = new MaxMinAS(stockSizes, itemSizes, stockNumbers, itemNumbers, param, rand);
		Solution solution = algoritam.findSolution(rand);
		
		PrintWriter writer = null;
		try {
			writer = new PrintWriter(new BufferedWriter(new FileWriter("rjesenja\\MaxMinAS_" + args[0] + ".out", true)));
		} catch (IOException e) {
			System.err.println("Greška pri otvaranju izlazne datoteke!");
			e.printStackTrace();
		}
		writer.println("##########");
		for (UsedStock stock : solution) {
			writer.println(stock);
		}
		writer.close();
	}
	
}
