/*************************************************************************
Yet Another Evolutionary Algorithm uz Local Search (YAEAuzLS)

Program koristi evolucijske strategije kao tehniku optimizacije pri
rjeavanju problema krojenja 1D (engl. cutting-stock problem ili CSP).
Takoer periodino primjenjuje varijantu lokalnog pretraivanja nad
jedinkama populacije u nastojanju poboljanja rjeenja.
                                                                       
Autor: Marin Mari                                                          
*************************************************************************/

#include <cstdio>
#include <cstring>
#include <cmath>
#include <ctime>
#include <vector>
#include <set>
#include <map>
#include <utility>
#include <algorithm>

#define UNDEF 0x3f3f3f3f

using namespace std;

const char* al_name = "YAEAuzLS";
const char* rjesenja_folder = "rjesenja\";
const char* algoritmi_folder = "algoritmi\";
char* task_name; // naziv zadatka

int s_pop = 4; // broj jedinki u poetnoj populaciji (bez djece)
int n_kids = 1; // broj djece svake jedinke pri mutaciji
int n_i_3ptswaps = 1; // broj trostrukih zamjena rezova pri mutaciji
int n_s_3ptswaps = 1; // broj trostrukih zamjena zaliha pri mutaciji
int f_ls = 1; // uestalost koritenja lokalnog pretraivanja (npr. 2 je svaki drugi ciklus mutacija selekcija)
int inf_stock_input = -1; // broj predstavlja beskonanu zalihu u ulaznoj datoteci

clock_t length_clocks = 10 * CLOCKS_PER_SEC; // broj taktova tj. vrijeme izvoenja algoritma
int cut_loss = 0; // gubitak zalihe pri rezanju ili debljina pile
int gd_cut = 0; // dovoljno dobar rez

// predstavlja jednu jedinku
struct entity
{
	double fitness; // dobrota rjeenja
	vector<int> stock; // zaliha
	vector<int> item; // rezovi
	bool need_ls; // potrebno lokalno pretraivanje (nije ve bilo)

	entity() {
		fitness = UNDEF;
		need_ls = true;
	}
};

entity e; // poetna jedinka
vector<entity*> pop; // populacija jedinki
int n_item = 0; // broj rezova u optjecaju
int n_stock = 0; // broj zaliha u optjecaju
map<int, pair<int, int> > inf_stock; // <beskonana zaliha, (broj istih konanih, pomona varijabla)>
int n_inf_stock = 0; // broj beskonanih zaliha u optjecaju

/*
Uitava postavke algoritma iz vanjske datoteke.
*/
void* load_config()
{
	char f_path[100];
	strcpy(f_path, algoritmi_folder);
	strcat(f_path, al_name);
	strcat(f_path, ".conf");
	FILE* f = fopen(f_path, "r");
	if (f == NULL)
		return NULL;

	char s[20];
	int x;

	while (fscanf(f, "%s%d", s, &x) != EOF) {
		if (!strcmp(s, "s_pop")) {
			s_pop = x;
		}
		else if (!strcmp(s, "n_kids")) {
			n_kids = x;
		}
		else if (!strcmp(s, "n_i_3ptswaps")) {
			n_i_3ptswaps = x;
		}
		else if (!strcmp(s, "n_s_3ptswaps")) {
			n_s_3ptswaps = x;
		}
		else if (!strcmp(s, "f_ls")) {
			f_ls = x;
		}
		else if (!strcmp(s, "inf_stock_input")) {
			inf_stock_input = x;
		}
	}

	fclose(f);
	return &e;
}

/*
Uitava ulaznu datoteku s opisom zadatka iz vanjske datoteke.
*/
void* load_input()
{
	char f_path[100];
	strcpy(f_path, algoritmi_folder);
	strcat(f_path, task_name);
	strcat(f_path, ".in");
	FILE* f = fopen(f_path, "r");
	if (f == NULL) {
		return NULL;
	}

	char d1, d2;
	int x, n;

	fscanf(f, "%d%d%d", &x, &cut_loss, &gd_cut);
	length_clocks = x * CLOCKS_PER_SEC;

	fseek(f, 1, SEEK_CUR);
	map<int, int> stock;
	while (fscanf(f, "%c%c", &d1, &d2) != EOF) {
		if (d1 == '-' || d2 == '-') {
			break;
		}
		fseek(f, -2, SEEK_CUR);
		fscanf(f, "%d%d", &x, &n);
		if (n == inf_stock_input) {
			inf_stock.insert(pair<int, pair<int, int> >(x, pair<int, int>(0, 0)));
		}
		else {
			stock.insert(pair<int, int>(x, n));
		}
		fseek(f, 1, SEEK_CUR);
	}

	do {
		if (fscanf(f, "%c", &d1) == EOF) {
			break;
		}
	} while (d1 == '-');
	fseek(f, -1, SEEK_CUR);

	while (fscanf(f, "%c%c", &d1, &d2) != EOF) {
		if (d1 == '-' || d2 == '-') {
			break;
		}
		fseek(f, -2, SEEK_CUR);
		fscanf(f, "%d%d", &x, &n);
		e.item.insert(e.item.end(), n, x);
		n_item += n;
		
		for (map<int, int>::iterator it = stock.begin(); it != stock.end(); ++it) {
			if (it->first >= x) {
				int cur_n = n;
				if (it->second < n) {
					cur_n = it->second;
				}
				if (it->second > 0) {
					e.stock.insert(e.stock.end(), cur_n, it->first);
					n_stock += cur_n;
					it->second -= cur_n;

					map<int, pair<int, int> >::iterator jt = inf_stock.find(it->first);
					if (jt != inf_stock.end()) {
						jt->second.first += cur_n;
					}
				}
			}
		}

		for (map<int, pair<int, int> >::iterator it = inf_stock.begin(); it != inf_stock.end(); ++it) {
			if (it->first >= x) {
				e.stock.insert(e.stock.end(), n, it->first);
				n_stock += n;
				n_inf_stock += n;
			}
		}

		fseek(f, 1, SEEK_CUR);
	}
	for (map<int, pair<int, int> >::iterator it = inf_stock.begin(); it != inf_stock.end(); ++it) {
		it->second.second = 0;
	}

	if (!n_inf_stock) {
		n_inf_stock = -1;
	}

	vector<int>(e.item).swap(e.item);
	vector<int>(e.stock).swap(e.stock);

	fclose(f);
   	return &e;
}

/*
Usporeuje dvije jedinke po dobroti, koriteno pri sortiranju populacije.
*/
bool cmp_entity_fitness(const entity* a, const entity* b)
{
	return (a->fitness < b->fitness);
}

/*
Postavlja dobrotu jedinke.
*/
void fitness_1(entity& x)
{
	int i = 0, j = 0, tj = 0, ti = 0, Tj, sum, unused = 0;
	x.fitness = 0;

	for (; i < n_item; ++j) {
		if (x.stock[j] < x.item[i]) {
			++unused;
			continue;
		}

		map<int, pair<int, int> >::iterator it = inf_stock.find(x.stock[j]);
		if (it != inf_stock.end()) {
			++(it->second.second);
		}

		sum = x.item[i++];
		for (; i < n_item && (sum + x.item[i] + cut_loss <= x.stock[j]); sum += (x.item[i++] + cut_loss));
		if ((Tj = x.stock[j] - sum) > gd_cut) {
			++tj;
		}

		x.fitness += sqrt((double) Tj / x.stock[j]);
	}

	j -= unused;
	for (map<int, pair<int, int> >::iterator it = inf_stock.begin(); it != inf_stock.end(); ++it) {
		if (it->second.second > it->second.first) {
			ti += (it->second.second - it->second.first);
		}
		it->second.second = 0;
	}

	x.fitness = (x.fitness + ((double) tj / j) + ((double) ti / n_inf_stock)) / (j + 2);
}

/*
Predstavlja genetski operator mutacije. Izvrava se nad itavom populacijom.
*/
void muta_pop_1 ()
{
	int i, j, k, tmp, o = s_pop;

	for (int n = 0; n < n_kids; ++n)
		for (int m = 0; m < s_pop; ++m) {
			pop[o]->stock.assign(pop[m]->stock.begin(), pop[m]->stock.end());
			pop[o]->item.assign(pop[m]->item.begin(), pop[m]->item.end());

			for (int l = 0; l < n_s_3ptswaps; ++l) {
				i = rand() % n_stock;
				do {
					j = rand() % n_stock;
				} while (j == i);
				do {
					k = rand() % n_stock;
				} while (k == i || k == j);
			
				tmp = pop[o]->stock[i];
				pop[o]->stock[i] = pop[o]->stock[k];
				pop[o]->stock[k] = pop[o]->stock[j];
				pop[o]->stock[j] = tmp;
			}

			for (int l = 0; l < n_i_3ptswaps; ++l) {
				i = rand() % n_item;
				do {
					j = rand() % n_item;
				} while (j == i);
				do {
					k = rand() % n_item;
				} while (k == i || k == j);
		
				tmp = pop[o]->item[i];
				pop[o]->item[i] = pop[o]->item[k];
				pop[o]->item[k] = pop[o]->item[j];
				pop[o]->item[j] = tmp;
			}

			pop[o]->need_ls = true;
			pop[o++]->fitness = UNDEF;
		}
}

/*
Predstavlja genetski operator selekcije. Nakon izvravanja jedinke su
poredane po dobroti unutar populacije.
*/
void sel_pop_1()
{
	for (int i = pop.size() - 1; i >= 0 && pop[i]->fitness == UNDEF; fitness_1(*pop[i--]));
	sort(pop.begin(), pop.end(), cmp_entity_fitness);
}

/*
Inicijalizira i postavlja populaciju u poetno stanje.
*/
void init_pop_1()
{
	pop.reserve(s_pop * (n_kids + 1));
	for (int i = 0; i < s_pop; ++i) {
		pop.push_back(new entity(e));
		random_shuffle(pop.back()->stock.begin(), pop.back()->stock.end());
		random_shuffle(pop.back()->item.begin(), pop.back()->item.end());
	}
	for (int i = 0; i < s_pop * n_kids; ++i) {
		pop.push_back(new entity(e));
	}
}

/*
Lokalno pretraivanje nad cijelom populacijom. Pokuava poboljati rjeenje
oslobaanjem zaliha (njihovih rezova) i pametnim ponovnim umetanjem u rjeenje.
*/
void ls_pop_1()
{
	int i, j, Tj, free_l = UNDEF, free_l_tmp, free_r = UNDEF, free_v = 0, free_j = UNDEF, sum;
	for (vector<entity*>::iterator pit = pop.begin(); pit < pop.end(); ++pit) {
		entity& x = **pit;
		if (!x.need_ls) {
			continue;
		}
		x.need_ls = false;
		x.fitness = UNDEF;

		// oslobaanje rezova najgore zalihe (s najvie gubitaka)
		i = 0; j = 0; free_v = 0;
		for (; i < (int) x.item.size(); ++j) {
			free_l_tmp = i;

			if (x.stock[j] < x.item[i]) {
				continue;
			}

			sum = x.item[i++];
			for (; i < (int) x.item.size() && (sum + x.item[i] + cut_loss <= x.stock[j]); sum += (x.item[i++] + cut_loss));
			if ((Tj = x.stock[j] - sum) <= gd_cut) {
				continue;
			}

			if (Tj > free_v) {
				free_j = j;
				free_v = Tj;
				free_l = free_l_tmp;
				free_r = i;
			}
		}

		if (free_v <= gd_cut) {
			return;
		}

		x.stock.push_back(x.stock[free_j]);
		x.stock.erase(x.stock.begin() + free_j);
		multiset<int> free_items(x.item.begin() + free_l, x.item.begin() + free_r);
		x.item.erase(x.item.begin() + free_l, x.item.begin() + free_r);

		// oslobaanje rezova jedne nasumine zalihe
		i = 0; j = 0; free_v = rand() % x.item.size(); free_j = UNDEF;
		for (; i < (int) x.item.size(); ++j) {
			if (x.stock[j] < x.item[i]) {
				continue;
			}

			free_l_tmp = i;

			sum = x.item[i++];
			for (; (i < (int) x.item.size()) && (sum + x.item[i] + cut_loss <= x.stock[j]); sum += (x.item[i++] + cut_loss));
			if ((x.stock[j] - sum) <= gd_cut) {
				continue;
			}
			
			free_j = j;
			free_l = free_l_tmp;
			free_r = i;
			
			if (free_l_tmp <= free_v && free_v < i) {
				free_j = j;
				free_l = free_l_tmp;
				free_r = i;
				break;
			}

			if (i > free_v) {
				break;
			}
		}
		
		if (free_j != UNDEF) {
			x.stock.push_back(x.stock[free_j]);
			x.stock.erase(x.stock.begin() + free_j);
			free_items.insert(x.item.begin() + free_l, x.item.begin() + free_r);
			x.item.erase(x.item.begin() + free_l, x.item.begin() + free_r);
		}

		// ponovno umetanje osloboenih rezova dok svi nisu umetnuti
		while(!free_items.empty()) {
			int t = *(--free_items.end());
			free_items.erase(--free_items.end());
			bool inserted = false;

			// umetanje ukoliko ima dovoljno mjesta u nekoj od zaliha
			for (j = 0, i = 0; i < (int) x.item.size(); ++j) {
				if (x.stock[j] < x.item[i]) {
					if (x.stock[j] < t) {
						continue;
					}
					x.item.insert(x.item.begin() + i, t);
					inserted = true;
					break;
				}

				sum = x.item[i++];
				for (; i < (int) x.item.size() && (sum + x.item[i] + cut_loss <= x.stock[j]); sum += (x.item[i++] + cut_loss));
			       
				if ((x.stock[j] - sum) >= (t + cut_loss)) {
					x.item.insert(x.item.begin() + i, t);
					inserted = true;
					break;
				}
			}

			if (inserted) {
				continue;
			}

			// umetanje slobodnog reza umjesto manjeg, manji je osloboen
			for (j = 0, i = 0; i < (int) x.item.size(); ++j) {
				if (x.stock[j] < x.item[i]) {
					continue;
				}

				int k = i;
				sum = x.item[i++];
				for (; i < (int) x.item.size() && (sum + x.item[i] + cut_loss <= x.stock[j]); sum += (x.item[i++] + cut_loss));
				if ((Tj = x.stock[j] - sum) <= gd_cut) {
					continue;
				}

				for (; k < i; ++k) {
					if (t > x.item[k] && (t - x.item[k] <= Tj)) {
						free_items.insert(x.item[k]);
						x.item[k] = t;
						inserted = true;
						break;
					}
				}

				if (inserted) {
					break;
				}
			}

			if (inserted) {
				continue;
			}

			// umetanje na staru zalihu ukoliko nigdje nije mogue bolje smjestiti rez
			x.item.push_back(t);
		}
	}
}

/*
Ispisuje rezultat (rjeenje) algoritma u vanjsku datoteku.
*/
void* write_result() {
	char f_path[100];
	strcpy(f_path, rjesenja_folder);
	strcat(f_path, al_name);
	strcat(f_path, "_");
	strcat(f_path, task_name);
	strcat(f_path, ".out");
	FILE* f = fopen(f_path, "a");
	if (f == NULL) {
		return NULL;
	}

	fprintf(f, "##########");
	
	entity& x = *pop.front();
	int i = 0, j = 0, sum;
	for (; i < n_item; ++j) {
		if (x.stock[j] < x.item[i]) {
			continue;
		}

		fprintf(f, "\n%d\t%d+", x.stock[j], x.item[i]);
		sum = x.item[i++];
		for (; i < n_item && (sum + x.item[i] + cut_loss <= x.stock[j]); sum += (x.item[i++] + cut_loss)) {
			fprintf(f, "%d+", x.item[i]);
		}
	}
	
	fprintf(f, "\n");
	fclose(f);
	return &e;
}

/*
Uitava postavke algoritma i zadatak. Inicijalizira populaciju i izvrava
cikluse mutacije, lokalnog pretraivanja i selekcije dok ne istekne vrijeme
ili dok nije postignuta savrena dobrota. Potom ispisuje rjeenje.
*/
int main(int argc, char** argv)
{
	clock_t total_clock = clock();

	srand(time(NULL));

	if (argc != 2) {
		fprintf(stderr, "\nInvalid arguments, expected only the output file name. Exiting ...\n");
		return 0;	
	}
	task_name = *(argv + 1);

	if (load_input() == NULL) {
		fprintf(stderr, "\nCould not open %s%s.in. Exiting ...\n", algoritmi_folder, task_name);
		return 0;
	}

	if (load_config() == NULL) {
		fprintf(stderr, "\nCould not open %s%s.conf. Exiting ...\n", rjesenja_folder, al_name);
		return 0;
	}

	init_pop_1();

	unsigned int counter = 0;
	while (((clock() - total_clock) <= length_clocks) && pop.front()->fitness > 0) {
		muta_pop_1();

		if (f_ls && !(counter % f_ls)) {
			ls_pop_1();
		}

		sel_pop_1();

		++counter;
	}

	if (write_result() == NULL) {
		fprintf(stderr, "\nCould not write result to %s%s_%s.out. Exiting ...\n", rjesenja_folder, al_name, task_name);
		return 0;
	}

	return 0;
}
